// Scintilla source code edit control
// PlatFOX.cpp - implementation of platform facilities on FOX toolkit
// The License.txt file describes the conditions under which this software may be distributed.
// SB: initial release; based on ScintillaWX by Robin Dunn.

#include <ctype.h>
#include <fox/fx.h>
#include "Platform.h"
#include "stc.h"


Point Point::FromLong(long lpoint) {
  return Point(lpoint & 0xFFFF, lpoint >> 16);    //SB ????
}

Point PointFromFXPoint(FXPoint pt) {
  return Point(pt.x, pt.y);
}

FXPoint FXPointFromPoint(Point pt) {
  FXPoint fxpt;
  fxpt.x = pt.x; fxpt.y = pt.y;
  return fxpt;
}


//SB
// PRectangle:   (left, top, right, bottom)
// FXRectangle:  (x, y, w, h)

//  +-----------------------------------------------------------> (X)
//  |
//  |
//  |        left       right            x = left
//  |         x                          y = top
//  |         +----------+ y top         w = right - left
//  |         |          |               h = bottom - top
//  |         |          |
//  |         |          | h             left   = x
//  |         |          |               top    = y
//  |         |          |               right  = x + w
//  |         +----------+   bottom      bottom = y + h
//  |               w
//  |
//  |
//  v (Y)



FXRectangle FXRectangleFromPRectangle(PRectangle prc) {
  FXRectangle rc;
  rc.x = prc.left; rc.y = prc.top;
  rc.w = prc.right-prc.left; rc.h = prc.bottom-prc.top;
  return rc;
}


PRectangle PRectangleFromFXRectangle(FXRectangle rc) {
  PRectangle prc(rc.x, rc.y, rc.w+rc.x, rc.h+rc.y);    //SB ?
  return prc;
}


Colour::Colour(long lcol) {
  co = FXRGB(lcol & 0xff, (lcol >> 8) & 0xff, (lcol >> 16) & 0xff);    //SB ?
}


Colour::Colour(unsigned int red, unsigned int green, unsigned int blue) {
  co = FXRGB(red, green, blue);    //SB ?
}


bool Colour::operator==(const Colour &other) const {
    return co == other.co;
}


long Colour::AsLong() const {
  return ((FXBLUEVAL(co) << 16) |
         (FXGREENVAL(co) << 8) |
         (FXREDVAL(co)));            //SB ?
}


unsigned int Colour::GetRed() {
  return FXREDVAL(co);
}

unsigned int Colour::GetGreen() {
  return FXGREENVAL(co);
}

unsigned int Colour::GetBlue() {
  return FXBLUEVAL(co);
}


Palette::Palette() {
    used = 0;
    allowRealization = false;
}

Palette::~Palette() {
    Release();
}

void Palette::Release() {
    used = 0;
}


// Needed?
// This method either adds a colour to the list of wanted colours (want==true)
// or retrieves the allocated colour back to the ColourPair.
// This is one method to make it easier to keep the code for wanting and retrieving in sync.
void Palette::WantFind(ColourPair &cp, bool want) {
    if (want) {
        for (int i=0; i < used; i++) {
            if (entries[i].desired == cp.desired)
                return;
        }

        if (used < numEntries) {
            entries[used].desired = cp.desired;
            entries[used].allocated = cp.desired;
            used++;
        }
    } else {
        for (int i=0; i < used; i++) {
            if (entries[i].desired == cp.desired) {
                cp.allocated = entries[i].allocated;
                return;
            }
        }
        cp.allocated = cp.desired;
    }
}

void Palette::Allocate(Window &) {
    if (allowRealization) {}
}


Font::Font() {
    id = 0;
}

Font::~Font() {
}

void Font::Create(const char *faceName, int characterSet, int size, bool bold, bool italic) {
  Release();

  FXApp* app = FXApp::instance();
  id = new FXFont(app,
                  faceName,                                        // name
                  size,                                            // size
                  bold ? FONTWEIGHT_BOLD : FONTWEIGHT_NORMAL,      // weight
                  italic ? FONTSLANT_ITALIC : FONTSLANT_REGULAR,   // slant
                  FONTENCODING_DEFAULT,                            // encoding
                  0);                                              // hints
  id->create();

}

void Font::Release() {
  if (id) delete id;
  id = 0;
}


// from Platform.h:
// typedef FXDrawable* SurfaceID;
// FXDrawable* drawable;
// FXImage* bitmap;
// FXDCWindow* hdc;

Surface::Surface() :
    drawable(0),
    hdc(0),
    bitmap(0),
    x(0), y(0),
    inited(false),
    createdDC(false) {
}


Surface::~Surface() {
    Release();
}


void Surface::Release() {
    drawable = 0;
    if (createdDC) {
      delete hdc;
      hdc = 0;
      createdDC = false;
    }

    if (bitmap) {
      delete bitmap;
      bitmap = 0;
    }

    x=0; y=0;
    inited = false;

}


bool Surface::Initialised() {
    return inited;
}

void Surface::Init() {
    Release();
    inited = true;
}


void Surface::Init(FXDrawable* drawable_) {
    Release();
    drawable = drawable_;
    hdc = new FXDCWindow(drawable_);
    createdDC = true;
    inited = true;
    //printf("Surface: Inited\n");
}


void Surface::InitPixMap(int width, int height, Surface *surface_) {
  //SB: what to do with surface_ ?
  //printf("PLatFOX: ...InitPixmap() request for BMP %d:%d\n", width, height);

  Release();
  FXApp* app = FXApp::instance();
  FXImage* bitmap = new FXImage(app, NULL, 0, width, height);
  bitmap->create();
  drawable = bitmap;
  hdc = new FXDCWindow(bitmap);    //<--------
  createdDC = true;
  inited = true;
}



void Surface::PenColour(Colour fore) {
  hdc->setForeground(fore.co);
}

void Surface::BrushColor(Colour back) {
  hdc->setBackground(back.co);
}

void Surface::SetFont(Font &font_) {
  if (font_.GetID()) {
    hdc->setTextFont(font_.GetID());
  }
}


int Surface::LogPixelsY() {
  return 72;                                      //from PlatGTK
}


int Surface::DeviceHeightFont(int points) {
	int logPix = LogPixelsY();
	return (points * logPix + logPix / 2) / 72;     //from PlatGTK
}


void Surface::MoveTo(int x_, int y_) {
  x = x_;
  y = y_;
}

void Surface::LineTo(int x_, int y_) {
  hdc->drawLine(x, y, x_, y_);
  x = x_;
  y = y_;
}


void Surface::Polygon(Point *pts, int npts, Colour fore,
                      Colour back) {
  PenColour(fore);
  BrushColor(back);
  hdc->fillPolygon((FXPoint* )pts, npts);
}


void Surface::RectangleDraw(PRectangle rc, Colour fore, Colour back) {
  PenColour(fore);
  BrushColor(back);
  FXRectangle rec = FXRectangleFromPRectangle(rc);
  hdc->drawRectangle(rec.x, rec.y, rec.w, rec.h);
}


void Surface::FillRectangle(PRectangle rc, Colour back) {
  PenColour(back);
  FXRectangle rec = FXRectangleFromPRectangle(rc);
  hdc->fillRectangle(rec.x, rec.y, rec.w, rec.h);
}


void Surface::FillRectangle(PRectangle rc, Surface &surfacePattern) {
  FXRectangle rec = FXRectangleFromPRectangle(rc);

  //printf("PlatFOX: FillRectangle(patt) rec= %d,%d,%d,%d\n",
  //       rec.x, rec.y, rec.w, rec.h);

  hdc->setFillStyle (FILL_TILED);
  hdc->setTile((FXImage*)surfacePattern.drawable, rec.x, rec.y);
  hdc->fillRectangle(rec.x, rec.y, rec.w, rec.h);
  hdc->setFillStyle (FILL_SOLID);
}


void Surface::RoundedRectangle(PRectangle rc, Colour fore, Colour back) {
  PenColour(fore);
  BrushColor(back);
  RectangleDraw(rc, fore, back);    //SB: for now...
}


void Surface::Ellipse(PRectangle rc, Colour fore, Colour back) {
  PenColour(fore);
  BrushColor(back);
  FXRectangle rec = FXRectangleFromPRectangle(rc);
  hdc->drawArc(rec.x, rec.y, rec.w, rec.h, 0, 32767);    //SB: numbers from PlatGTK
}


void Surface::Copy(PRectangle rc, Point from, Surface &surfaceSource) {
  FXRectangle rec = FXRectangleFromPRectangle(rc);

  // note the args sequence! WHY??
  //printf("PlatFOX: Surface::Copy() from_size= %d,%d,%d,%d, to= %d,%d\n",
  //       from.x, from.y, rec.w, rec.h,    rec.x, rec.y);

  hdc->setFunction(BLT_SRC);
  hdc->drawArea(surfaceSource.drawable, from.x, from.y, rec.w, rec.h,
                                        rec.x, rec.y);
}


void Surface::DrawText(PRectangle rc, Font &font, int ybase,
                       const char *s, int len, Colour fore, Colour back) {
  BrushColor(back);
  FillRectangle(rc, back);
  SetFont(font);
  PenColour(fore);
  hdc->drawText(rc.left, ybase, s, len);
  ////hdc->drawImageText(rec.x, ybase, s, len);

  //printf ("PlatFOX Surface::DrawText() pt= (%d %d %d %d), ybase= %d, len= %d\n",
  //        rec.x, rec.y, rec.w, rec.h, ybase, len);
}


void Surface::DrawTextClipped(PRectangle rc, Font &font, int ybase, const char *s, int len, Colour fore, Colour back) {
  // needs testing!
  SetFont(font);
  PenColour(fore);
  BrushColor(back);
  FXRectangle rec = FXRectangleFromPRectangle(rc);
  hdc->setClipRectangle(rec.x, rec.y, rec.w, rec.h);
  hdc->drawImageText(rec.x, rec.y, s, len);
  hdc->clearClipRectangle();

  //printf("PlatFOX DrawTextClipped()\n");
}



int Surface::WidthText(Font &font, const char *s, int len) {
  return (*font.GetID()).getTextWidth(s, len);
}


void Surface::MeasureWidths(Font &font, const char *s, int len, int *positions) {
  int totalWidth = 0;
  for (int i=0; i<len; i++) {
    int w = (*font.GetID()).getTextWidth(&s[i], 1);
    totalWidth += w;
    positions[i] = totalWidth;
  }
}


int Surface::WidthChar(Font &font, char ch) {
  return (*font.GetID()).getTextWidth(&ch, 1);
}


#define EXTENT_TEST " `~!@#$%^&*()-_=+\\|[]{};:\"\'<,>.?/1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"


int Surface::Ascent(Font &font) {
  return (*font.GetID()).getFontAscent();
}


int Surface::Descent(Font &font) {
  return (*font.GetID()).getFontDescent();
}


int Surface::InternalLeading(Font &font) {
    return 0;    //SB ?
}


int Surface::ExternalLeading(Font &font) {
  return (*font.GetID()).getFontLeading();    //SB ?
}


int Surface::Height(Font &font) {
  return (*font.GetID()).getFontHeight();
}

int Surface::AverageCharWidth(Font &font) {
  return (*font.GetID()).getFontWidth();
}

int Surface::SetPalette(Palette *pal, bool inBackGround) {
    return 0;  //SB ??
}


void Surface::SetClip(PRectangle rc) {
  FXRectangle rec = FXRectangleFromPRectangle(rc);
  hdc->setClipRectangle(rec.x, rec.y, rec.w, rec.h);
}

void Surface::FlushCachedState() {
}




Window::~Window() {
}

void Window::Destroy() {
    if (id) id->destroy();
    id = 0;
}

bool Window::HasFocus() {
  return id->hasFocus();
}


PRectangle Window::GetPosition() {
  FXRectangle rec;
  rec.x = id->getX(); rec.y = id->getY();
  rec.w = id->getDefaultWidth(); rec.h = id->getDefaultHeight();  //or getWidth()?

  //printf("PlatFOX: GetPosition() called\n");
  return PRectangleFromFXRectangle(rec);    //SB ?
}


void Window::SetPosition(PRectangle rc) {
  FXRectangle rec = FXRectangleFromPRectangle(rc);
  id->position(rec.x, rec.y, rec.w, rec.h);
}


void Window::SetPositionRelative(PRectangle rc, Window) {
  SetPosition(rc);  // ????
}


PRectangle Window::GetClientPosition() {
// warning, she wants the size of the viewport area!!!!

  FXRectangle rec;
  rec.x = 0; rec.y = 0;
  rec.w = ((FXStyledText *)id)->getViewportWidth();
  rec.h = ((FXStyledText *)id)->getViewportHeight();
  PRectangle rc = PRectangleFromFXRectangle(rec);

  //printf("PlatFOX: GetClientPosition() w= %d, h= %d\n", rec.w, rec.h);
  return rc;
}


void Window::Show(bool show) {
  id->show();
}


void Window::InvalidateAll() {
  //printf("PlatformFOX: InvalidateAll() called\n");
  //id->forceRefresh();
  id->update();
}


void Window::InvalidateRectangle(PRectangle rc) {
  FXRectangle rec = FXRectangleFromPRectangle(rc);
  id->update(rec.x, rec.y, rec.w, rec.h);
}


void Window::SetFont(Font &font) {
  //FXApp* app = FXApp::instance();
  //app->setNormalFont(font.GetID());   //??

  FXDCWindow hdc((FXScrollArea*)id);    //FXDrawable??
  hdc.setTextFont(font.GetID());
}

void Window::SetCursor(Cursor curs) {
//printf("PlatFOX: SetCursor() called: ");

    FXDefaultCursor cursorId;

    switch (curs) {
    case cursorText:
        cursorId = DEF_TEXT_CURSOR;          //printf("CURSOR_IBEAM\n");
        break;
    case cursorArrow:
        cursorId = DEF_ARROW_CURSOR;         //printf("CURSOR_ARROW\n");
        break;
    case cursorUp:
        cursorId = DEF_ARROW_CURSOR;         //printf("CURSOR_ARROW\n");
        break;
    case cursorWait:
        cursorId = DEF_SWATCH_CURSOR;        //printf("CURSOR_WATCH\n");
        break;
    case cursorHoriz:
        cursorId = DEF_HSPLIT_CURSOR;        //printf("CURSOR_LEFTRIGHT\n");
        break;
    case cursorVert:
        cursorId = DEF_VSPLIT_CURSOR;        //printf("CURSOR_UPDOWN\n");
        break;
    case cursorReverseArrow:
        //cursorId = DEF_RARROW_CURSOR;      //printf("CURSOR_RARROW\n");
				cursorId = DEF_ARROW_CURSOR;         // I don't like the rarrow
        break;
    default:
        cursorId = DEF_ARROW_CURSOR;         //printf("CURSOR_ARROW\n");
        break;
    }

		id->setDefaultCursor(id->getApp()->getDefaultCursor(cursorId));
}


void Window::SetTitle(const char *s) {
    //id->SetTitle(s);    //SB ????
}


/////////////////
//Todo Listbox...

/*
From WxScintilla:

class wxSTCListBox : public wxListBox {
public:
    wxSTCListBox(wxWindow* parent, wxWindowID id)
        : wxListBox(parent, id, wxDefaultPosition, wxDefaultSize,
                    0, NULL, wxLB_SINGLE | wxLB_SORT | wxSIMPLE_BORDER)
        {}

    void OnFocus(wxFocusEvent& event) {
        GetParent()->SetFocus();
        event.Skip();
    }

#ifdef __WXGTK__
    void DoSetFirstItem(int n);
#endif

private:
    DECLARE_EVENT_TABLE()
};

BEGIN_EVENT_TABLE(wxSTCListBox, wxListBox)
    EVT_SET_FOCUS(wxSTCListBox::OnFocus)
END_EVENT_TABLE()


#ifdef __WXGTK__
    // This can be removed after 2.2.2 I think
void wxSTCListBox::DoSetFirstItem( int n )
{
    wxCHECK_RET( m_list, wxT("invalid listbox") );

    if (gdk_pointer_is_grabbed () && GTK_WIDGET_HAS_GRAB (m_list))
        return;

    // terribly efficient
    const gchar *vadjustment_key = "gtk-vadjustment";
    guint vadjustment_key_id = g_quark_from_static_string (vadjustment_key);

    GtkAdjustment *adjustment =
       (GtkAdjustment*) gtk_object_get_data_by_id (GTK_OBJECT (m_list), vadjustment_key_id);
    wxCHECK_RET( adjustment, wxT("invalid listbox code") );

    GList *target = g_list_nth( m_list->children, n );
    wxCHECK_RET( target, wxT("invalid listbox index") );

    GtkWidget *item = GTK_WIDGET(target->data);
    wxCHECK_RET( item, wxT("invalid listbox code") );

    // find the last item before this one which is already realized
    size_t nItemsBefore;
    for ( nItemsBefore = 0; item && (item->allocation.y == -1); nItemsBefore++ )
    {
        target = target->prev;
        if ( !target )
        {
            // nothing we can do if there are no allocated items yet
            return;
        }

        item = GTK_WIDGET(target->data);
    }

    gtk_adjustment_set_value(adjustment,
                             item->allocation.y +
                                nItemsBefore*item->allocation.height);
}
#endif
*/


ListBox::ListBox() {
}

ListBox::~ListBox() {
}

void ListBox::Create(Window &parent, int ctrlID) {
    //id = new wxSTCListBox(parent.id, ctrlID);
}

PRectangle ListBox::GetDesiredRect() {
    //wxSize sz = ((wxListBox*)id)->GetBestSize();
  PRectangle rc;
    //rc.top = 0;
    //rc.left = 0;
    //if (sz.x > 150)   // TODO: A better way to determine these max sizes
    //    sz.x = 150;
    //if (sz.y > 100)
    //    sz.y = 100;
    //rc.right = sz.x;
    //rc.bottom = sz.y;
  return rc;
}

void ListBox::SetAverageCharWidth(int width) {
    //aveCharWidth = width;
}

void ListBox::SetFont(Font &font) {
    //Window::SetFont(font);
}

void ListBox::Clear() {
    //((wxListBox*)id)->Clear();
}

void ListBox::Append(char *s) {
    //((wxListBox*)id)->Append(s);
}

int ListBox::Length() {
    //return ((wxListBox*)id)->Number();
}

void ListBox::Select(int n) {
//    ((wxListBox*)id)->SetSelection(n);
//#ifdef __WXGTK__
//    if (n > 4)
//        n = n - 4;
//    else
//        n = 1;
//    ((wxListBox*)id)->SetFirstItem(n);
//#endif
}

int ListBox::GetSelection() {
    //return ((wxListBox*)id)->GetSelection();
}

int ListBox::Find(const char *prefix) {
    //if (prefix) {
    //    for (int x=0; x < ((wxListBox*)id)->Number(); x++) {
    //        wxString text = ((wxListBox*)id)->GetString(x);
    //        if (text.StartsWith(prefix))
    //            return x;
    //    }
    //}
    return -1;
}

void ListBox::GetValue(int n, char *value, int len) {
    //wxString text = ((wxListBox*)id)->GetString(n);
    //strncpy(value, text.c_str(), len);
    //value[len-1] = '\0';
}

void ListBox::Sort() {
    // wxWindows keeps sorted so no need to sort
}
/////////////////


// SB: Scintilla PopUp not used
Menu::Menu() : id(0) {}
void Menu::CreatePopUp() {}
void Menu::Destroy() {}
void Menu::Show(Point pt, Window &w) {}


//SB from FXApp.cpp
#define borderColor  FXRGB(0,0,0)
#define baseColor    FXRGB(192,192,192)
#define hiliteColor  makeHiliteColor(baseColor)
#define shadowColor  makeShadowColor(baseColor)
#define backColor    FXRGB(255,255,255)
#define foreColor    FXRGB(0,0,0)


Colour Platform::Chrome() {
  return baseColor;
}


Colour Platform::ChromeHighlight() {
  return hiliteColor;
}


const char *Platform::DefaultFont() {
  FXFont* deffont = FXApp::instance()->getNormalFont();
  FXString stName = deffont->getName();
  const char *name = stName.text();

  //printf("PlatFOX: DefaultFont() called, returning [%s]\n", name);
  return name;
}


int Platform::DefaultFontSize() {
  int size = FXApp::instance()->getNormalFont()->getSize();
  //printf("PlatFOX: DefaultFontSize() called, returning %d\n", size/10);

  return size/10;  //why?
}


unsigned int Platform::DoubleClickTime() {
    return 500;
}


void Platform::DebugDisplay(const char *s) {
  //SB ??
}


bool Platform::IsKeyDown(int key) {
    return false;  // I don't think we'll need this.  //SB ??
}


long Platform::SendScintilla(WindowID w,
                             unsigned int msg,
                             unsigned long wParam,
                             long lParam) {

    FXStyledText* stc = (FXStyledText* )w;
    return stc->SendMsg(msg, wParam, lParam);
}


// These are utility functions not really tied to a platform

int Platform::Minimum(int a, int b) {
    if (a < b) return a;
    else return b;
}


int Platform::Maximum(int a, int b) {
    if (a > b) return a;
    else return b;
}


#define TRACE

void Platform::DebugPrintf(const char *format, ...) {
int TRACE_LEVEL=100;
#ifdef TRACE
    char buffer[2000];
    va_list pArguments;
    va_start(pArguments, format);
    vsprintf(buffer,format,pArguments);
    va_end(pArguments);
    //Platform::DebugDisplay(buffer);

  FXTRACE((TRACE_LEVEL, buffer));    //SB: Note double parents!
#endif
}


int Platform::Clamp(int val, int minVal, int maxVal) {
    if (val > maxVal) val = maxVal;
    if (val < minVal) val = minVal;
    return val;
}


void Platform::Assert(const char *c, const char *file, int line) {
  printf("ASSERT: %s, FILE: %s, LINE: %d\n", c, file, line);
}



